/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/


/*	AEP_I_Submit:
	This function submits a command to the AEP Daemon without waiting on the reply (Asynchronous).
	The daemon returns a transaction handle for this command which is used to retrieve the reply.

	Inputs	:	Connection Handle  {hConnection}
				Pointer to the Message Block (Header and data) {pCommandBuf}
				Pointer to a Transaction handle (either a valid ptr or NULL)  {phTransHndl}

	Outputs	:	Transaction handle is returned if a valid pointer was passed in phTransHndl
*/
AEP_RV AEP_I_Submit(
        AEP_CONNECTION_HNDL		hConnection,
		AEP_BYTEBLOCK*			pCommandBuf
      );
      
/*	AEP_I_Transact:
	This function submits a command to the AEP Daemon and waits for the reply (Synchronous).
	
	Inputs	:	Connection Handle  {hConnection}
				Pointer to the Message Block (Header and data) {pCommandBuf}
				Pointer to the Reply Block (Header and data)  {pReplyBlock}

	Outputs	:	The Reply is returned in the reply block.
*/


AEP_RV AEP_I_Transact(
        AEP_CONNECTION_HNDL hConnection,
		AEP_BYTEBLOCK*		pCommandBuf,
		AEP_BYTEBLOCK*		pReplyBuf
      );
      
      
/*	AEP_I_Query:
	This function checks to see if a result has become available.
	
	Inputs	:	Connection Handle  {hConnection}
				Transaction Handle {hTransaction}
				Pointer to the Reply Block (Header and data)  {pReplyBlock}

	Outputs	:	The Reply is returned in the reply block.
*/
AEP_RV AEP_I_Query(
        AEP_CONNECTION_HNDL hConnection,
		AEP_BYTEBLOCK*		pCommandBuf,
		AEP_BYTEBLOCK*		pReplyBuf
      );
      
      
/*	AEP_I_Wait:
	This function waits for aresult to become available.
	
	Inputs	:	Connection Handle  {hConnection}
				Transaction Handle {hTransaction}
				Pointer to the Reply Block (Header and data)  {pReplyBlock}

	Outputs	:	The Reply is returned in the reply block.
*/
AEP_RV AEP_I_Wait(
        AEP_CONNECTION_HNDL hConnection,
		AEP_BYTEBLOCK*		pCommandBuf,
		AEP_BYTEBLOCK*		pReplyBuf
      );
	  
/*	AEP_I_GetDaemonVersion:
	This function gets the version number of the daemon software.
	
	Inputs	:	Connection Handle  {hConnection}
				Pointer to the Command block {pCommandBuf}
				Pointer to the Reply Block (Header and data)  {pReplyBuf}

	Outputs	:	The Reply is returned in the reply block.
*/
AEP_RV AEP_I_GetDaemonVersion(
        AEP_CONNECTION_HNDL hConnection,
	AEP_BYTEBLOCK*		pCommandBuf,
	AEP_BYTEBLOCK*		pReplyBuf 
	);      
