/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/



#ifndef _MUTEX_H_
#define _MUTEX_H_ 1

#include <assert.h>
#include <semaphore.h>

#include "aep_i_types.h"
#include "aep_i_defs.h"

AEP_RV AEP_I_CreateMutex(
  AEP_VOID_PTR_PTR ppMutex  /* location to receive ptr to mutex */
);

AEP_RV AEP_I_DestroyMutex(
  AEP_VOID_PTR pMutex  /* ptr to mutex */
);

AEP_RV AEP_I_LockMutex(
  AEP_VOID_PTR pMutex  /* ptr to mutex */
);

AEP_RV AEP_I_UnlockMutex(
  AEP_VOID_PTR pMutex  /* ptr to mutex */
);

int SEM_INIT(sem_t *sem, int pshared, unsigned int value);

int SEM_WAIT(sem_t *arg);

#ifdef NDEBUG
#define _LOCK(__mutex)       (AEP_I_LockMutex( __mutex ))
#define _UNLOCK(__mutex)     (AEP_I_UnlockMutex( __mutex ))
#else 
#define _LOCK(__mutex)       assert(AEP_I_LockMutex( __mutex ) == AEP_R_OK)
#define _UNLOCK(__mutex)     assert(AEP_I_UnlockMutex( __mutex ) == AEP_R_OK)
#endif

#endif /* _MUTEX_H_ */
