/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

/* OS specific headers */
#ifdef AEP_Win32
#include <windows.h>
#include <process.h>
#elif defined AEP_GENERIC
#include <pthread.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>

#include "aeptrace.h"
#include "aep_i_defs.h"
#include "aep_i_types.h"
#include "aep_api.h"
#include "mutex.h"


/*************************************************************************************
   aeptrace.c 
	
   This file contains functions which implement system tracing 
   The user supplies required tracing parameters via a configuration
   file: aeplog.conf
 
************************************************************************************/
/* Global Variables */
/* Mutex */
static pthread_mutex_t file_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Holds user defined configuration options */
static TRACE_CONFIG_ITEM TRACE_OPTIONS[NUM_TRACE_CONFIG_ITEMS];
AEP_U32 setTraceRV;
/******************************************************************************************************/
/*
  void AEP_TRACE (AEP_U32		Subsystem,		 Subsystem calling trace i.e API, Daemon or Driver 
	    	  AEP_U32			TraceLevel,		 Message importance
			  AEP_U32			Indentation,	 Function in, Function out or Function none 	
			  AEP_CHAR_PTR		FunctionName,	 Name of function
			  AEP_RV			rv,				 Operation Result
			  AEP_CHAR_PTR		Message,		 Logging Message 
			  ...)								 Variable length message string	

*/
/******************************************************************************************************/
void AEP_TRACE (AEP_U32			Subsystem,
				AEP_U32			TraceLevel,
				AEP_U32			Indentation,
				AEP_CHAR_PTR	FunctionName,
				AEP_U32			rv,
				AEP_CHAR_PTR	Message,
				...
				)			
{
	static int firstTime = 0;
	va_list params;
		
	/* If this is the first time trace has been called read the
	   configuration file and initialise all variables
	*/
	if(firstTime == 0)
	{
		init();
		firstTime = 1;
	}/* end if */	
		
	/* Check that tracing is on for this subsystem */
	if((Subsystem == API) && (TRACE_OPTIONS[AEP_API_TRACE_STATUS].Status == OFF))
	{
		return;
	}
	else if((Subsystem == DAEMON) && (TRACE_OPTIONS[AEP_DAEMON_TRACE_STATUS].Status == OFF))
	{
		return;
	}
	else if(TRACE_OPTIONS[TraceLevel].Status == ON)
	{
		/* Ensure message isn't null */
		Message = (Message == (void*)NULL) ? "(NULL)" : Message;
	
		/* Copy variable length possibly formatted string with arguments into buffer */	
		va_start(params, Message);
		
		/* Log Message*/
		logMessage(Subsystem,TraceLevel,Indentation,FunctionName,rv,Message,params);
	}
				
	return;
}
/******************************************************************************************************/	
/*
  void logMessage (AEP_U32		Subsystem,		 Subsystem calling trace i.e API, Daemon or Driver 
	    	  AEP_U32			TraceLevel,		 Message importance
			  AEP_U32			Indentation,	 Function in, Function out or Function none 	
			  AEP_CHAR_PTR		FunctionName,	 Name of function
			  AEP_RV			rv,				 Operation Result
			  AEP_CHAR_PTR		Message,		 Logging Message 
			  va_list		    params			Variable length message string
			  )								 

*/
/******************************************************************************************************/	
void logMessage(AEP_U32			Subsystem,
				AEP_U32			TraceLevel,
				AEP_U32			Indentation,
				AEP_CHAR_PTR	FunctionName,
				AEP_U32			rv,
				AEP_CHAR_PTR	Message,
				va_list			params
				
				)				
{			
	AEP_CHAR subsysLogfile[MAX_FILE_LENGTH] = {""};
	AEP_CHAR_PTR startString;
	AEP_CHAR_PTR currentTimeString;
	AEP_CHAR_PTR tempTimeString;
	const char* fieldSpace = "     ";
	AEP_U32 ProcessID;
	AEP_U32 ThreadID;
	char buffer[MAX_TRACE_LENGTH ];
	int bufferIndex = 0;
	static int initFile = 0;
	time_t currentTime;
	int stringLength;
	int tempInt;
	FILE *SUBSYS_LOG;
	int i;
	static time_t startTime;
	const char* delim = "\n";
		
	/* Ensure function name is never null */
    FunctionName = (FunctionName == NULL) ? "(NULL)" : FunctionName;
		
	/* Ensure Indentation is never null */
	if((Indentation != FN_IN) && (Indentation != FN_OUT) && (Indentation != FN_NONE))
		Indentation = FN_NONE;
		
	/* Set the log file to use based on the calling system */
	/* File names are set in aeptrace.h */
	switch(Subsystem)
	{
		case DAEMON:
			/* Prepend logging directory defined by user if one exists*/
			strncpy(subsysLogfile, TRACE_OPTIONS[DAEMON_LOG_DIR].name, TRACE_OPTIONS[DAEMON_LOG_DIR].length);
			/* Add file name to directory string */
			strcat(subsysLogfile, DAEMON_LOG_FILE);
			/* Lock Daemon File */
   	        _LOCK(&file_mutex);
			break;
		case API:
			/* Prepend logging directory defined by user if one exists*/
		 	strncpy(subsysLogfile, TRACE_OPTIONS[API_LOG_DIR].name, TRACE_OPTIONS[API_LOG_DIR].length);
			/* Add file name to directory string */
			strcat(subsysLogfile, API_LOG_FILE);
			/* Lock API File */	
  	        _LOCK(&file_mutex);
			break;
		case DRIVER:
			/* Add file name to directory string */
			strncpy(subsysLogfile, TRACE_OPTIONS[DRIVER_LOG_DIR].name, TRACE_OPTIONS[DRIVER_LOG_DIR].length);
			/* Add file name to directory string */
			strcat(subsysLogfile, DRIVER_LOG_FILE);
			/* Lock Driver File */
	        _LOCK(&file_mutex);
			break;
	}	/* end switch */

	/* If filelogging is enabled and file has not yet been initialised, initialise it
	   with filename and date
	*/
	if(initFile == 0)
	{
		startTime = time(NULL);
		/* ctime results in a time string ending in \n\0 - strtok results in a substring up until the 
		   specified character is reached, in this case `\n`.  Basically it returns the time string 
		   minus the \n\0
		*/  
		startString = ctime(&startTime);
		startString = (AEP_CHAR_PTR)strtok(startString, delim);
			
		/* Create the log file - if a file of that name
		   exists, destroy contents
		*/

		if(TRACE_OPTIONS[FILE_OUTPUT_ENABLED].Status == ON)
		{
			if((SUBSYS_LOG = fopen(subsysLogfile, "w")) == NULL)
			{
				perror("\nError Opening Log File \n");
				_UNLOCK(&file_mutex);
				rv = AEP_R_FILE_OPEN_ERROR;
				return;
			}/* end if */

			/* Write Header to File */	
			fprintf(SUBSYS_LOG,"*******************************\n%s\n%s\n*******************************\n",subsysLogfile,startString);

			initFile = 1;
		}/* end if */
	}/* end if */
	/* File has already been initialised so open in append mode */
	else if(TRACE_OPTIONS[FILE_OUTPUT_ENABLED].Status == ON)
	{	
		/* Open file in append mode */
		if((SUBSYS_LOG = fopen(subsysLogfile, "a")) == NULL)
		{
			perror("\nError Opening Log File \n");
			_UNLOCK(&file_mutex);
			rv = AEP_R_FILE_OPEN_ERROR;
			return;
		}/* end if */
	}/* end else */

	/* If Timestamp option set, get current time */
	if(TRACE_OPTIONS[TIME_STAMP_ENABLED].Status == ON)
	{
		currentTime = time(NULL);	
		currentTimeString = ctime(&currentTime);
		tempTimeString = (AEP_CHAR_PTR)strtok(currentTimeString, delim);
		stringLength = strlen(tempTimeString);
			
		/* Copy time into buffer */
		sprintf(&buffer[bufferIndex], "%s",tempTimeString);
		bufferIndex += stringLength;
		tempInt = RelTimeWidth - stringLength;
		for(i = 0; i < tempInt; i++)
			sprintf(&buffer[bufferIndex+i]," ");
		bufferIndex += tempInt;
	}
	/* If show process ID option set copy the process ID into the buffer */
	if(TRACE_OPTIONS[SHOW_PROCESS_ID].Status  == ON)
	{
		/* Get the Process ID */
		ProcessID = getpid();
		/* Ensure process ID is not null */
		stringLength = sprintf(&buffer[bufferIndex],"[PID : %ul]%s", ProcessID, fieldSpace);
		bufferIndex += stringLength;
		tempInt = PIDWidth - stringLength;
		for(i = 0; i < tempInt; i++)
			sprintf(&buffer[bufferIndex+i]," ");
		bufferIndex += tempInt;
	}/* end if */

	/* If show thread ID option set copy the thread ID into the buffer */
	if(TRACE_OPTIONS[SHOW_THREAD_ID].Status == ON)
	{
		/* Get the Thread and Process ID */
		ThreadID = pthread_self();

		
		/* Ensure thread ID is not null */
		stringLength = sprintf(&buffer[bufferIndex],"[TID : %ul]%s", ThreadID, fieldSpace);
		bufferIndex += stringLength;
		tempInt = TIDWidth - stringLength;
		for( i = 0; i < tempInt; i++)
			sprintf(&buffer[bufferIndex+i]," ");
		bufferIndex += tempInt;
	}/* end if */

	/* Prefix each logging message with function Name */
	stringLength = sprintf(&buffer[bufferIndex],"%s%s", FunctionName, fieldSpace);
	bufferIndex += stringLength;
	tempInt = FNNameWidth - stringLength;
	for(i = 0; i < tempInt; i++)
		sprintf(&buffer[bufferIndex+i]," ");
	bufferIndex += tempInt;
	
	/* Based on Indentation specified, add an extra line indicating function entry point */
	if(Indentation == FN_IN)
	{
		/* Add space to account for lack of rv */
		for(i = 0; i < RVWidth; i++)
		{
			stringLength=sprintf(&buffer[bufferIndex]," ");
			bufferIndex++;
		}
		stringLength = sprintf(&buffer[bufferIndex],"Enter %s%s\n", FunctionName, fieldSpace);
		bufferIndex+=stringLength;

		/* If timestamp enabled prefix message with timestamp */
		if(TRACE_OPTIONS[TIME_STAMP_ENABLED].Status == ON)
		{
			stringLength = sprintf(&buffer[bufferIndex],"%s%s", tempTimeString, fieldSpace);
			bufferIndex+= stringLength;
			tempInt = RelTimeWidth - stringLength;
			for(i = 0; i < tempInt; i++)
				sprintf(&buffer[bufferIndex+i]," ");
			bufferIndex += tempInt;
		}	
		/* If show process ID option set add process ID to buffer*/
		if(TRACE_OPTIONS[SHOW_PROCESS_ID].Status == ON)
		{
			stringLength = sprintf(&buffer[bufferIndex],"[PID : %ul]%s", ProcessID, fieldSpace);
			bufferIndex += stringLength;
			tempInt = PIDWidth - stringLength;
			for( i = 0; i < tempInt; i++)
				sprintf(&buffer[bufferIndex+i]," ");
			bufferIndex += tempInt;
		}/* end if */
		
		/* If show thread ID option set add thread ID to buffer */
		if(TRACE_OPTIONS[SHOW_THREAD_ID].Status == ON)
		{
			stringLength = sprintf(&buffer[bufferIndex],"[TID : %ul]%s",ThreadID, fieldSpace);
			bufferIndex += stringLength;
			tempInt = TIDWidth - stringLength;
			for( i = 0; i < tempInt; i++)
				sprintf(&buffer[bufferIndex+i]," ");
			bufferIndex += tempInt;
		}/* end if */
			
		/* Add function Name */
		stringLength = sprintf(&buffer[bufferIndex],"%s%s", FunctionName, fieldSpace);
		bufferIndex += stringLength;
		tempInt = FNNameWidth - stringLength;
		for(i = 0; i < tempInt; i++)
			sprintf(&buffer[bufferIndex+i]," ");
		bufferIndex += tempInt;
	}/* end if */
	
	/* Add RV */
	if(rv > 0)
	{	
		stringLength = sprintf(&buffer[bufferIndex], "[RV : 0x%0x]%s", rv, fieldSpace);
		bufferIndex += stringLength;
		tempInt = RVWidth - stringLength;
		for(i = 0; i < tempInt; i++)
			sprintf(&buffer[bufferIndex+i]," ");
		bufferIndex += tempInt;
	}
	else
	{
		stringLength = sprintf(&buffer[bufferIndex], "[RV : N/A]%s", fieldSpace);
		bufferIndex += stringLength;
		tempInt = RVWidth - stringLength;
		for(i = 0; i < tempInt; i++)
			sprintf(&buffer[bufferIndex+i]," ");
		bufferIndex += tempInt;
	}	

	/* Add message to buffer */
	stringLength = vsprintf(&buffer[bufferIndex], Message,params);
	bufferIndex += stringLength;

	/* Finished with variable length parameters */
	va_end(params);
		

	/* If this is the last log of a function postfix a line indicating function exit point */
	if (Indentation == FN_OUT)
	{
		sprintf(&buffer[bufferIndex],"\n");
		bufferIndex++;
		/* If timestamp enabled copy timestamp into buffer */
		if(TRACE_OPTIONS[TIME_STAMP_ENABLED].Status == ON)
		{
			stringLength = sprintf(&buffer[bufferIndex],"%s%s", tempTimeString, fieldSpace);
			bufferIndex +=stringLength;
			tempInt = RelTimeWidth - stringLength;
			for(i = 0; i < tempInt; i++)
				sprintf(&buffer[bufferIndex+i]," ");
			bufferIndex += tempInt;
				
		}/*end if */	
			
		/* If show process ID option set add process ID to buffer*/
		if(TRACE_OPTIONS[SHOW_PROCESS_ID].Status == ON)
		{
			stringLength = sprintf(&buffer[bufferIndex],"[PID : %ul]%s",ProcessID, fieldSpace);
			bufferIndex += stringLength;
			tempInt = PIDWidth - stringLength;
			for( i = 0; i < tempInt; i++)
				sprintf(&buffer[bufferIndex+i]," ");
			bufferIndex += tempInt;
		}/* end if */
		/* If show thread ID option set add thread ID to buffer */
		if(TRACE_OPTIONS[SHOW_THREAD_ID].Status == ON)
		{
			stringLength = sprintf(&buffer[bufferIndex],"[TID : %ul]%s", ThreadID, fieldSpace);
			bufferIndex += stringLength;
			tempInt = TIDWidth - stringLength;
			for(i = 0; i < tempInt; i++)
				sprintf(&buffer[bufferIndex+i]," ");
			bufferIndex += tempInt;
		}/* end if */
		/* Add function Name */
		stringLength = sprintf(&buffer[bufferIndex],"%s%s", FunctionName, fieldSpace);
		bufferIndex += stringLength;
		tempInt = FNNameWidth - stringLength;
		for(i = 0; i < tempInt; i++)
			sprintf(&buffer[bufferIndex+i]," ");
		bufferIndex += tempInt;

		/* Add space to account for lack of rv */
		for(i = 0; i < RVWidth; i++)
		{
			stringLength=sprintf(&buffer[bufferIndex]," ");
			bufferIndex++;
		}
		/* Add exit string */
		stringLength = sprintf(&buffer[bufferIndex], "Exit %s%s",FunctionName, fieldSpace);
		bufferIndex+=stringLength;
	}

	/* Write newline character to buffer */
    sprintf(&buffer[bufferIndex], "\n");
	bufferIndex++;

	/* If log is to be written to console -print it*/
	if(TRACE_OPTIONS[STREAM_OUTPUT_ENABLED].Status == ON)
		fputs(buffer, stdout);	
	/* If log is to be written to file */
	if(subsysLogfile != NULL)
	{
		if(TRACE_OPTIONS[FILE_OUTPUT_ENABLED].Status == ON)
		{
			/* Write buffer to file */
			fputs((AEP_CHAR_PTR)buffer, SUBSYS_LOG);	
			/* Close file */
			fclose(SUBSYS_LOG);
		}	
		/* Unlock mutex on appropriate file */
		_UNLOCK(&file_mutex);
	}/* end if */
	return;
}/* end AEPTrace */
/******************************************************************************************************/
/* Function called on library load */
void init(void)
{	
	AEP_RV rv;
#ifdef AEP_GENERIC
	struct sigaction sigAct;
#endif	
	/* Initialise all trace items to off */
	rv = initTraceConfigItems();

	/* Parse config file and set configurable options  */
	setTraceConfigValues(&setTraceRV);
	
	if (setTraceRV != AEP_R_OK)
	{
		printf("\nsetTraceRV not ok\n");
		perror("Error configuring variables\n");
		exit(1);
	}
	
	/*Catch SIGUSR1 signal that the management application sends us this tell us that
	the configuration variables have changed and we should reload them */
#ifdef AEP_GENERIC		
	sigAct.sa_handler = setTraceConfigValues;
	sigemptyset(&sigAct.sa_mask);
	sigAct.sa_flags = SA_RESTART;
	
	if (sigaction(SIGTTOU, &sigAct, NULL) < 0)
	{	
		perror("Can't catch SIGTTOU signals\n");
	}/* end if */
#endif

}/* end main */
/******************************************************************************************************
	int initTraceConfigItems()	

	This function reads the tracing options required by the user from aeplog.conf and populates 
	the global array of structures with these values 
******************************************************************************************************/
int initTraceConfigItems()
{
	int i;
	
	for(i = 0; i < NUM_TRACE_CONFIG_ITEMS; i++)
	{
		TRACE_OPTIONS[i].item = i;
		TRACE_OPTIONS[i].Status = OFF;
		TRACE_OPTIONS[i].length = 0;
	}/* end for */

	return AEP_R_OK;
}/* end initTraceConfigItems */
/******************************************************************************************************
	int setConfigItem(AEP_U32 item			Identifies configuration item to set
					  AEP_U32 status		ON / OFF
					  AEP_CHAR_PTR string	Value may be a string such as a filename
					  AEP_U32 length		Length of value string if it exists
	)	

	This function reads the tracing options required by the user from aeplog.conf and populates 
	the global array of structures with these values 
******************************************************************************************************/
AEP_U32 setConfigItem(AEP_U32 item, AEP_U32 status, AEP_CHAR* string, AEP_U32 length)
{
	/* If item exists */
	if((item >= 0) && (item < NUM_TRACE_CONFIG_ITEMS))
	{	
		/* Set the status */
		TRACE_OPTIONS[item].Status = status;	
		/* If a valid string has been passed i.e. string length > 0; set the string */
		if(length > 0)
		{
			TRACE_OPTIONS[item].length = length;
			strncpy(TRACE_OPTIONS[item].name, string, length);
		}
	}	/* end if */
	return AEP_R_OK;
}
/******************************************************************************************************/
