/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/
#include <pthread.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include "aep_i_defs.h"
#include "aep_i_types.h"
#include "aep_daemon_defs.h"
#include "config.h"
#include "aependian.h"
#include "sockets.h"
#include "aepmalloc.h"
#include "aeptrace.h"

#include "mutex.h"

#include "buf_hndlr.h"


/*This is where the requests are buffered while waiting to be sent to the accelerator*/
extern AEP_U64 ASIC_RequestBuffer[2][BUFFER_SIZE/8];
extern AEP_U64 DES_RequestBuffer[2][BUFFER_SIZE/8];
extern AEP_U64 SA_RequestBuffer[2][BUFFER_SIZE/8];

extern AEP_U64* ASIC_CurRequestBuffer;
extern AEP_U64* DES_CurRequestBuffer;
extern AEP_U64* SA_CurRequestBuffer;

/*Accelerator File descriptor*/
extern int AccFd;

/*Offset into request Buffers*/
extern AEP_U32 ASICBuffIndex;
extern AEP_U32 DESBuffIndex;
extern AEP_U32 SABuffIndex;

/*Used with the condition variable when deciding if buffer should be moved to accelerator*/
extern AEP_U32 SendASICBuffer;
extern AEP_U32 SendDESBuffer;
extern AEP_U32 SendSABuffer;

/*This synchronisation variables are used to control access to shared data*/
extern pthread_mutex_t ASIC_RequestBuffer_mutex;
extern pthread_mutex_t DES_RequestBuffer_mutex;
extern pthread_mutex_t SA_RequestBuffer_mutex;

/*Condition variables used to wait for the request buffers to be sent/to need sending*/
extern pthread_cond_t ASIC_send_now, ASIC_send_complete;
extern pthread_cond_t DES_send_now, DES_send_complete;
extern pthread_cond_t SA_send_now, SA_send_complete;

extern pthread_cond_t ASIC_BufferFull_cond;
extern pthread_cond_t DES_BufferFull_cond;
extern pthread_cond_t SA_BufferFull_cond;

extern AEP_U32 alive_message_flag;
extern pthread_mutex_t AEP_ContextID_mutex;

/*	ptrASICBuffHndlr
	This function defines the thread that handles the emptying of the ASIC request buffer
*/
void *ptrASICBuffHndlr()
{
	int rv, bad_write;
	
	AEP_U64* buf_ptr;
	
	AEP_U32 cur_buf = 0;
	
	AEP_U32 buf_len;

	
  /*These two structs are used to calculate the timeout for the condition variable*/
	struct timespec timeout;
	struct timeval timenow; 

  /* Sleep on a condition variable and when woken check to see if there's any data,
   if there is send it, signal the original thread and continue*/

/*Lock the request buffer mutex*/
	_LOCK(&ASIC_RequestBuffer_mutex);
	
  /*Check the request buffer forever*/
	while(1)
	{

	/*This is the condition that is evaluated for the condition variable.  So before
		another thread signals this condition variable it changes this condition.  In this way
		we can ignore spurious wake ups and other concurrency probs.
			If Send Buffer is true then go ahead and send it otherwise keep waiting*/
		while (!SendASICBuffer)
		{
		  /*When specifying how long the timeout period for the condition variable is
			an absolute time must be used.  So we must get the current time and add the
			timeout interval to that time*/
			gettimeofday(&timenow, NULL);
			timeout.tv_sec = timenow.tv_sec;

		  /*time is in nano seconds so this equates to current time plus 10 milli secs*/
			timeout.tv_nsec = timenow.tv_usec * 1000 + ASIC_WriteTimeout;


		  /*If the last calculation resulted in an overflow then add one to the secs
			and decrease the nanosecs by 1 secs worth of nano secs*/
			if (timeout.tv_nsec > 1000000000)
			{
				timeout.tv_sec++;
				timeout.tv_nsec -= 1000000000;
			}

			/*Wait for the condition variable to be signalled or time out
				Note: This automatically releases the mutex.  On return we automatically
				reacquire the mutex*/
			rv = pthread_cond_timedwait(&ASIC_send_now, &ASIC_RequestBuffer_mutex, &timeout);
		  /*If we have returned because of a timeout*/
			if (rv == ETIMEDOUT)
			{
			  /*Only bother to send anything if we have anything to send*/
				if (ASICBuffIndex > 0)
					SendASICBuffer = TRUE;
			}
		}

		/*Remember the buf ptr and the buf len*/
		buf_ptr = &ASIC_RequestBuffer[cur_buf][0];
		buf_len = ASICBuffIndex;

		cur_buf++;
		if(cur_buf == 2) cur_buf=0;
		
	   /*Swap the buffer pointers*/
	  	ASIC_CurRequestBuffer = &ASIC_RequestBuffer[cur_buf][0];

	  /*Reset the BuffIndex (Effectively emptying the buffer)*/
		ASICBuffIndex = 0;

	  /*Set the SendBuffer variable to FALSE*/
		SendASICBuffer = FALSE;

	  /*Send a signal to any threads that are waiting to add requests to the request buffer*/
		pthread_cond_broadcast(&ASIC_send_complete);
		
/*UnLock the request buffer mutex*/
		_UNLOCK(&ASIC_RequestBuffer_mutex);	
		
		bad_write=0;

	  /*Write the block of messages to the target*/		
		rv = write(AccFd, buf_ptr, buf_len);

/*Lock the request buffer mutex*/
		_LOCK(&ASIC_RequestBuffer_mutex);

	  /*Check the write*/		
		while(rv < 0)
		{
			if (errno == EAGAIN)
			{
			  /*When specifying how long the timeout period for the condition variable is
				an absolute time must be used.  So we must get the current time and add the
				timeout interval to that time*/
				
				gettimeofday(&timenow, NULL);
				timeout.tv_sec = timenow.tv_sec;

			  /*time is in nano seconds so this equates to current time plus 10 milli secs*/
				timeout.tv_nsec = timenow.tv_usec * 1000 + ASIC_WriteTimeout;

			  /*If the last calculation resulted in an overflow then add one to the secs
				and decrease the nanosecs by 1 secs worth of nano secs*/
				if (timeout.tv_nsec > 1000000000)
				{
					timeout.tv_sec++;
					timeout.tv_nsec -= 1000000000;
				}

				rv = pthread_cond_timedwait(&ASIC_BufferFull_cond, &ASIC_RequestBuffer_mutex, &timeout);
			}
			else
			{
			  /*inc the number of bad writes*/
				bad_write++;
			}

			if (bad_write==10)
			{
				AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "ptrASICBuffHndlr", AEP_R_TARGET_ERROR, "Unable to Write to Accelerator - Exiting");
				perror("AEP bad write\n");
				exit(1);
			}
			
			rv = write(AccFd, buf_ptr, buf_len);
		}			
					
	  /*Insert log entry*/

		AEP_TRACE (DAEMON, TRACE_DEBUG, FN_NONE, "ptrASICBuffHndlr", AEP_R_OK, 
					"ASIC Request BufferHandler Thread sent %d Bytes to the Accelerator.", buf_len);


#ifdef AEP_CTX_ON			
/*Lock the ContextID mutex*/
		_LOCK(&AEP_ContextID_mutex);
	  /*Set this flag to indicate there's no need to send an alive message from the
	  	context hndlr thread 'cause we've just sent some requests*/
		alive_message_flag = 1;

/*UnLock the ContextID mutex*/	
		_UNLOCK(&AEP_ContextID_mutex);
#endif
		
	}
}

/*	ptrDESBuffHndlr
	This function defines the thread that handles the emptying of the DES request buffer
*/
void *ptrDESBuffHndlr()
{
	int rv, bad_write;

	AEP_U64* buf_ptr;
	
	AEP_U32 cur_buf = 0;
	
	AEP_U32 buf_len;


  /*These two structs are used to calculate the timeout for the condition variable*/
	struct timespec timeout;
	struct timeval timenow; 

  /* Sleep on a condition variable and when woken check to see if there's any data,
   if there is send it, signal the original thread and continue*/

/*Lock the request buffer mutex*/
	_LOCK(&DES_RequestBuffer_mutex);

  /*Check the request buffer forever*/
	while(1)
	{
		/*This is the condition that is evaluated for the condition variable.  So before
		another thread signals this condition variable it changes this condition.  In this way
		we can ignore spurious wake ups and other concurrency probs.
			If Send Buffer is true then go ahead and send it otherwise keep waiting*/
		while (!SendDESBuffer)
		{
		  /*When specifying how long the timeout period for the condition variable is
			an absolute time must be used.  So we must get the current time and add the
			timeout interval to that time*/
			gettimeofday(&timenow, NULL);
			timeout.tv_sec = timenow.tv_sec;

		  /*time is in nano seconds so this equates to current time plus 10 milli secs*/
			timeout.tv_nsec = timenow.tv_usec * 1000 + DES_WriteTimeout;

		  /*If the last calculation resulted in an overflow then add one to the secs
			and decrease the nanosecs by 1 secs worth of nano secs*/
			if (timeout.tv_nsec > 1000000000)
			{
				timeout.tv_sec++;
				timeout.tv_nsec -= 1000000000;
			}

			/*Wait for the condition variable to be signalled or time out
				Note: This automatically releases the mutex.  On return we automatically
				reacquire the mutex*/
			rv = pthread_cond_timedwait(&DES_send_now, &DES_RequestBuffer_mutex, &timeout);
		  /*If we have returned because of a timeout*/
			if (rv == ETIMEDOUT)
			{
			  /*Only bother to send anything if we have anything to send*/
				if (DESBuffIndex > 0)
					SendDESBuffer = TRUE;
			}
		}

		/*Remember the buf ptr and the buf len*/
		buf_ptr = &DES_RequestBuffer[cur_buf][0];
		buf_len = DESBuffIndex;

		cur_buf++;
		if(cur_buf == 2) cur_buf=0;
		
	   /*Swap the buffer pointers*/
	  	DES_CurRequestBuffer = &DES_RequestBuffer[cur_buf][0];

	  /*Reset the BuffIndex (Effectively emptying the buffer)*/
		DESBuffIndex = 0;

	  /*Set the SendBuffer variable to FALSE*/
		SendDESBuffer = FALSE;

	  /*Send a signal to any threads that are waiting to add requests to the request buffer*/
		pthread_cond_broadcast(&DES_send_complete);
		
/*UnLock the request buffer mutex*/
		_UNLOCK(&DES_RequestBuffer_mutex);

		bad_write=0;
		
	  /*Write the block of messages to the target*/		
		rv = write(AccFd, buf_ptr, buf_len);

/*Lock the request buffer mutex*/
		_LOCK(&DES_RequestBuffer_mutex);
		
	  /*Check the write*/		
		while(rv < 0)
		{
			if (errno == EAGAIN)
			{
			  /*When specifying how long the timeout period for the condition variable is
				an absolute time must be used.  So we must get the current time and add the
				timeout interval to that time*/
				
				gettimeofday(&timenow, NULL);
				timeout.tv_sec = timenow.tv_sec;

			  /*time is in nano seconds so this equates to current time plus 10 milli secs*/
				timeout.tv_nsec = timenow.tv_usec * 1000 + DES_WriteTimeout;

			  /*If the last calculation resulted in an overflow then add one to the secs
				and decrease the nanosecs by 1 secs worth of nano secs*/
				if (timeout.tv_nsec > 1000000000)
				{
					timeout.tv_sec++;
					timeout.tv_nsec -= 1000000000;
				}
				
				rv = pthread_cond_timedwait(&DES_BufferFull_cond, &DES_RequestBuffer_mutex, &timeout);
			}
			else
			{
			  /*inc the number of bad writes*/
				bad_write++;
			}

			if (bad_write==10)
			{
			
				AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "ptrDESBuffHndlr", AEP_R_TARGET_ERROR, "Unable to Write to Accelerator - Exiting");
				exit(1);
			}
			
			rv = write(AccFd, buf_ptr, buf_len);
		}			
		
	  /*Insert log entry*/
    	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_NONE, "ptrDESBuffHndlr", AEP_R_OK, 
					"DES Request BufferHandler Thread sent %d Bytes to the Accelerator.", buf_len);


#ifdef AEP_CTX_ON			
/*Lock the ContextID mutex*/
		_LOCK(&AEP_ContextID_mutex);
	  /*Set this flag to indicate there's no need to send an alive message from the
	  	context hndlr thread 'cause we've just sent some requests*/
		alive_message_flag = 1;

/*UnLock the ContextID mutex*/	
		_UNLOCK(&AEP_ContextID_mutex);
#endif
		
	}
}


/*	ptrSABuffHndlr
	This function defines the thread that handles the emptying of the SA request buffer
*/
void *ptrSABuffHndlr()
{
	int rv, bad_write;
	
	AEP_U64* buf_ptr;
	
	AEP_U32 cur_buf = 0;
	
	AEP_U32 buf_len;


  /*These two structs are used to calculate the timeout for the condition variable*/
	struct timespec timeout;
	struct timeval timenow; 
	
  /* Sleep on a condition variable and when woken check to see if there's any data,
   if there is send it, signal the original thread and continue*/

/*Lock the request buffer mutex*/
	_LOCK(&SA_RequestBuffer_mutex);

  /*Check the request buffer forever*/
	while(1)
	{

		/*This is the condition that is evaluated for the condition variable.  So before
		another thread signals this condition variable it changes this condition.  In this way
		we can ignore spurious wake ups and other concurrency probs.
			If Send Buffer is true then go ahead and send it otherwise keep waiting*/

		while (!SendSABuffer)
		{
		  /*When specifying how long the timeout period for the condition variable is
			an absolute time must be used.  So we must get the current time and add the
			timeout interval to that time*/
			gettimeofday(&timenow, NULL);
			timeout.tv_sec = timenow.tv_sec;

		  /*time is in nano seconds so this equates to current time plus 10 milli secs*/
			timeout.tv_nsec = timenow.tv_usec * 1000 + SA_WriteTimeout;

		  /*If the last calculation resulted in an overflow then add one to the secs
			and decrease the nanosecs by 1 secs worth of nano secs*/
			if (timeout.tv_nsec > 1000000000)
			{
				timeout.tv_sec++;
				timeout.tv_nsec -= 1000000000;
			}

			/*Wait for the condition variable to be signalled or time out
				Note: This automatically releases the mutex.  On return we automatically
				reacquire the mutex*/
			rv = pthread_cond_timedwait(&SA_send_now, &SA_RequestBuffer_mutex, &timeout);
		
		  /*If we have returned because of a timeout*/
			if (rv == ETIMEDOUT)
			{
			  /*Only bother to send anything if we have anything to send*/
				if (SABuffIndex > 0)
					SendSABuffer = TRUE;
			}
		}

		/*Remember the buf ptr and the buf len*/
		buf_ptr = &SA_RequestBuffer[cur_buf][0];
		buf_len = SABuffIndex;

		cur_buf++;
		if(cur_buf == 2) cur_buf=0;
		
	   /*Swap the buffer pointers*/
	  	SA_CurRequestBuffer = &SA_RequestBuffer[cur_buf][0];

	  /*Reset the BuffIndex (Effectively emptying the buffer)*/
		SABuffIndex = 0;

	  /*Set the SendBuffer variable to FALSE*/
		SendSABuffer = FALSE;

	  /*Send a signal to any threads that are waiting to add requests to the request buffer*/
		pthread_cond_broadcast(&SA_send_complete);

 	  /*UnLock the request buffer mutex*/
		_UNLOCK(&SA_RequestBuffer_mutex);

		bad_write=0;
		
	  /*Write the block of messages to the target*/		
		rv = write(AccFd, (AEP_U8*) buf_ptr, buf_len);

/*Lock the request buffer mutex*/
		_LOCK(&SA_RequestBuffer_mutex);
		
	  /*Check the write*/		
		while(rv < 0)
		{
			if (errno == EAGAIN)
			{
			  /*When specifying how long the timeout period for the condition variable is
				an absolute time must be used.  So we must get the current time and add the
				timeout interval to that time*/
				
				gettimeofday(&timenow, NULL);
				timeout.tv_sec = timenow.tv_sec;

			  /*time is in nano seconds so this equates to current time plus 10 milli secs*/
				timeout.tv_nsec = timenow.tv_usec * 1000 + SA_WriteTimeout;

			  /*If the last calculation resulted in an overflow then add one to the secs
				and decrease the nanosecs by 1 secs worth of nano secs*/
				if (timeout.tv_nsec > 1000000000)
				{
					timeout.tv_sec++;
					timeout.tv_nsec -= 1000000000;
				}

				rv = pthread_cond_timedwait(&SA_BufferFull_cond, &SA_RequestBuffer_mutex, &timeout);
			}
			else
			{
			  /*inc the number of bad writes*/
				bad_write++;
			}

			if (bad_write==10)
			{
				AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "ptrSABuffHndlr", AEP_R_TARGET_ERROR, "Unable to Write to Accelerator - Exiting");
				perror("AEP bad write\n");
				exit(1);
			}
			
			rv = write(AccFd, buf_ptr, buf_len);
		}			

	  /*Insert log entry*/

		AEP_TRACE (DAEMON, TRACE_DEBUG, FN_NONE, "ptrSABuffHndlr", AEP_R_OK, 
					"SA Request BufferHandler Thread sent %d Bytes to the Accelerator.", buf_len);


#ifdef AEP_CTX_ON			
/*Lock the ContextID mutex*/
		_LOCK(&AEP_ContextID_mutex);
	  /*Set this flag to indicate there's no need to send an alive message from the
	  	context hndlr thread 'cause we've just sent some requests*/
		alive_message_flag = 1;

/*UnLock the ContextID mutex*/	
		_UNLOCK(&AEP_ContextID_mutex);
#endif
	
	}
}
