/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/
/*
	AEP Daemon process 7th November Diarmuid O'Neill
	Header file ...aepsrv.h
*/
#include "aep_daemon_defs.h"
#include "../../h/aep_i_types.h"
#include "../../h/aep_i_defs.h"


/*******************Function Declerations*****************/

/*Function definition for the connection handler thread*/
static void *ptrConnHndlr(void*);

/*These next two functions just seperate daemon/device commands adding to the readability of the code*/
AEP_RV DoDaemonCommand(AEP_MSG_HDR_PTR message_hdr, int socket_fd);

AEP_RV DoDeviceCommand(AEP_MSG_HDR_PTR message_hdr, int socket_fd);

AEP_RV DoContextCommand(AEP_MSG_HDR_PTR message_hdr, int socket_fd);

AEP_RV DoKeyManCommand(AEP_MSG_HDR_PTR message_hdr, int socket_fd);


/************************Global Data***********************/

/*This is where the requests are buffered while waiting to be sent to the accelerator*/
/*Declared as arrays of U64's to stop alignment warnings*/
AEP_U64 ASIC_RequestBuffer[2][BUFFER_SIZE / 8];
AEP_U64 DES_RequestBuffer[2][BUFFER_SIZE / 8];
AEP_U64 SA_RequestBuffer[2][BUFFER_SIZE / 8];
AEP_U64* ASIC_CurRequestBuffer = &ASIC_RequestBuffer[0][0];
AEP_U64* DES_CurRequestBuffer = &DES_RequestBuffer[0][0];
AEP_U64* SA_CurRequestBuffer = &SA_RequestBuffer[0][0];


/*Offset into request Buffers*/
AEP_U32 ASICBuffIndex = 0;
AEP_U32 DESBuffIndex = 0;
AEP_U32 SABuffIndex = 0;

/*Used with the condition variable when deciding if buffer should be moved to accelerator*/
AEP_U32 SendASICBuffer = FALSE;
AEP_U32 SendDESBuffer = FALSE;
AEP_U32 SendSABuffer = FALSE;

/*Accelerator File descriptor*/
int AccFd;

/*Key Manager socket descriptor*/
int keydb_fd;

/*This synchronisation variables are used to control access to shared data*/
pthread_mutex_t ASIC_RequestBuffer_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t DES_RequestBuffer_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t SA_RequestBuffer_mutex = PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t AEP_KeyMan_mutex = PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t response_item_mutex_array[MAX_AEP_RESPONSE_TABLE_SIZE];

pthread_mutex_t AEP_ContextID_mutex = PTHREAD_MUTEX_INITIALIZER;

pthread_cond_t request_contexts = PTHREAD_COND_INITIALIZER;

/*Condition variables used to wait for the request buffers to be sent/to need sending*/
pthread_cond_t ASIC_send_now = PTHREAD_COND_INITIALIZER;
pthread_cond_t ASIC_send_complete = PTHREAD_COND_INITIALIZER;
pthread_cond_t DES_send_now = PTHREAD_COND_INITIALIZER;
pthread_cond_t DES_send_complete = PTHREAD_COND_INITIALIZER;
pthread_cond_t SA_send_now = PTHREAD_COND_INITIALIZER;
pthread_cond_t SA_send_complete = PTHREAD_COND_INITIALIZER;

pthread_cond_t ASIC_BufferFull_cond = PTHREAD_COND_INITIALIZER;
pthread_cond_t DES_BufferFull_cond = PTHREAD_COND_INITIALIZER;
pthread_cond_t SA_BufferFull_cond = PTHREAD_COND_INITIALIZER;

/*Semaphore used to limit the amount of connections the daemon accepts.*/
sem_t connection_counter;

/*Semaphore used to hold off taking requests before the daemon authenticates
	and requests some contexts*/
sem_t auth_semaphore;

/*Semaphore that the response handler thread waits on before it reads the
	done array on the target*/
sem_t read_done_array_semaphore;

/*This structure holds the responses hash table*/
AEP_I_SRV_DATA AEP_I_response_table = {NULL_PTR};

/*This variable holds the daemons process id, which is used to make up tran id's*/
AEP_U32 proc_ID;

/*This variable holds the deamons context ID (App ID) that is sent with
every request for more contexts*/
AEP_U64 daemon_context_id;

AEP_U64 contextID_array[NUM_CONTEXTS_REQUESTED + CONTEXTS_LOW_MARK + 1];

AEP_U32 contextID_count=0;

AEP_U32 alive_message_flag=0;
