/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/
#ifndef _HASH_H_
#define _HASH_H_ 1


#include "aep_i_defs.h"
#include "aep_i_types.h"

/* This is the hash function used to determine the index into the table from the key
   It's a very simple hash function but the sequential nature of the transaction handles
   compliments this hash function.  There are no possible collisions until the connection
   handle value is greater than the table size.*/

/*#define AEP_I_HASH(_table,_key)  ((_key) % (_table->tab_size)) */

/*Replaced previous hash routine with a more effecient version.  Note that the
number of entries in the has table MUST be a multiple of 2*/
/*#define AEP_I_HASH(_key)	 ((_key) & (AEP_HASH_MASK)) */
/* Replaced previous #define of AEP_I_HASH with a function, this is needed
   to make the dimension of the response table configurable */


typedef struct AEP_I_HASH_BUCKET {
AEP_VOID_PTR val;
AEP_U64 key;
AEP_U32 index;     /* index of this bucket in the top table. i.e. the result of the the hash function */
struct AEP_I_HASH_BUCKET * next;
} AEP_I_HASH_BUCKET;

typedef AEP_I_HASH_BUCKET * AEP_I_HASH_BUCKET_PTR;
typedef AEP_I_HASH_BUCKET_PTR * AEP_I_HASH_BUCKET_PTR_PTR;

typedef struct AEP_I_HASHTABLE {
AEP_U32 tab_size;  /* table size in entries */
AEP_U32 entries;   /* current number of entries */
AEP_I_HASH_BUCKET_PTR_PTR table;
} AEP_I_HASHTABLE;

typedef AEP_I_HASHTABLE * AEP_I_HASHTABLE_PTR;
typedef AEP_I_HASHTABLE_PTR * AEP_I_HASHTABLE_PTR_PTR;

AEP_U32 AEP_I_HASH ( AEP_U64 key);

extern AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_I_InitHashtable(
  AEP_I_HASHTABLE_PTR_PTR ppHashTable,  /* returned ptr to hash table */
  AEP_U32 size                     /* size of table to be used */
);

extern AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_I_DestroyHashtable(
  AEP_I_HASHTABLE_PTR pHashTable  /* hash table to be destroyed */
);

AEP_RV AEP_I_HashPutEntry(
  AEP_I_HASHTABLE_PTR pHashTable,  /* returned ptr to hash table */
  AEP_U64 key,       /* key of data, base for hash */
  AEP_VOID_PTR val             /* data to be inserted */
);

AEP_RV AEP_I_HashGetEntry(
  AEP_I_HASHTABLE_PTR pHashTable,  /* returned ptr to hash table */
  AEP_U64 key,                /* key of data, base for hash */
  AEP_VOID_PTR_PTR val         /* returned pointer to data*/
);

AEP_RV AEP_I_HashRemoveEntry(
  AEP_I_HASHTABLE_PTR pHashTable,  /* returned ptr to hash table */
  AEP_U64 key                /* key of data to be deleted */
);

AEP_RV AEP_I_HashEntryExists(
  AEP_I_HASHTABLE_PTR pHashTable,  /* returned ptr to hash table */
  AEP_U64 key                /* key of data to be deleted */
);

AEP_CHAR_PTR AEP_I_HashtableToString(
  AEP_I_HASHTABLE_PTR pHashTable  /* returned ptr to hash table */
);

/** calculate a good hash value over a string.
 * @param string string for which to compute the hashvalue
 */
AEP_U32 AEP_I_string_hash(
  AEP_CHAR_PTR string
);


typedef struct AEP_I_HASH_ITERATOR
{
  AEP_I_HASHTABLE_PTR table;
  AEP_I_HASH_BUCKET_PTR curr_bucket;
}AEP_I_HASH_ITERATOR;

typedef AEP_I_HASH_ITERATOR * AEP_I_HASH_ITERATOR_PTR;

/* pIterator is set to NULL_PTR if there is no element. */
/*
 * Warning: any change in the Hashtable may invalidate the Iterator!
 */
AEP_RV AEP_I_HashIterateInit(
 AEP_I_HASHTABLE_PTR pHashTable,
 AEP_I_HASH_ITERATOR_PTR * ppIterator
);

/* pIterator is set to NULL_PTR if there is no further element. */
AEP_RV AEP_I_HashIterateInc(
 AEP_I_HASH_ITERATOR_PTR pIterator
);

/** Derefence element of the hashtable from an iterator.
 * If return of either of key or val is not requiered, set pointer to 
 * NULL_PTR 
 */
AEP_RV AEP_I_HashIterateDeRef(
 AEP_I_HASH_ITERATOR_PTR pIterator,
 AEP_U64_PTR pKey,
 AEP_VOID_PTR_PTR ppVal
);

/** Delete the element of the hashtable the iterator points at.
 * will set the iterator to the next element in the hashtable or NULL_PTR 
 * if there is no further element.
 * The function compensates for the fact that a delete changes the
 * hash table.
 */
AEP_RV AEP_I_HashIterateDel(
 AEP_I_HASH_ITERATOR_PTR pIterator
);

/* check the state of the iterator */
AEP_BBOOL AEP_I_HashIterValid(
 AEP_I_HASH_ITERATOR_PTR pIterator
);

/* clear the iterator */
AEP_RV AEP_I_HashIterateDelete(
 AEP_I_HASH_ITERATOR_PTR pIterator
);


/*AEP_RV AEP_I_NewHandle(
  AEP_WORD_PTR handle*/                /* handle returned */
/*);*/

typedef struct AEP_I_SRV_DATA {
  AEP_I_HASHTABLE_PTR response_table;    /* responses table, holds responses */
} AEP_I_SRV_DATA;
extern AEP_I_SRV_DATA AEP_I_response_table;


#endif /* _HASH_H_ */
