/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/
#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#include "aep_i_defs.h"
#include "aep_i_types.h"
#include "aep_daemon_defs.h"
#include "config.h"
#include "keyman_hndlr.h"

#include "aepmalloc.h"

#include "mutex.h"
#include "hash.h"
#include "aeptrace.h"
#include "sockets.h"
#include "aependian.h"

extern int keydb_fd;

/*This is where the requests are buffered while waiting to be sent to the accelerator*/
extern AEP_U64* ASIC_CurRequestBuffer;
extern AEP_U64* DES_CurRequestBuffer;
extern AEP_U64* SA_CurRequestBuffer;


/*Offset into request Buffers*/
extern AEP_U32 ASICBuffIndex;
extern AEP_U32 DESBuffIndex;
extern AEP_U32 SABuffIndex;

/*Used with the condition variable when deciding if buffer should be moved to accelerator*/
extern AEP_U32 SendASICBuffer;
extern AEP_U32 SendDESBuffer;
extern AEP_U32 SendSABuffer;

/*This synchronisation variables are used to control access to shared data*/
extern pthread_mutex_t ASIC_RequestBuffer_mutex;
extern pthread_mutex_t DES_RequestBuffer_mutex;
extern pthread_mutex_t SA_RequestBuffer_mutex;

extern pthread_mutex_t response_item_mutex_array[];

extern pthread_mutex_t AEP_KeyMan_mutex;

/*Condition variables used to wait for the request buffers to be sent/to need sending*/
extern pthread_cond_t ASIC_send_now, ASIC_send_complete;
extern pthread_cond_t DES_send_now, DES_send_complete;
extern pthread_cond_t SA_send_now, SA_send_complete;


/*Memory mapped addresses*/
extern AEP_U8 *config_base_addr;

void *ptrKeyManHndlr()
{
	AEP_RV rv= AEP_R_OK;
	AEP_MSG_HDR message_hdr;
	
	AEP_I_RESPONSE_DATA_PTR pResponse;
	
	AEP_U32 i;

  /*Used to point to the RequestBuffer Mutex for this request*/
	pthread_mutex_t *AEP_RequestBuffer_mutex;

  /*Used to point to the condition variables for the sending/waiting to be sent of this buffer*/
	pthread_cond_t *send_now, *send_complete;

  /*Used to point to the offset into the request Buffer for this request*/
	AEP_U32 *BuffIndex;

  /*Used to point to the variable used when deciding if this buffer should go to the accelerator*/
	AEP_U32 *SendBuffer;

  /*Used to point to the Request Buffer for this request*/
	volatile AEP_U64 *AEP_RequestBuffer;

  /*Used to hold the size of the Request buffer associated with this request*/
	AEP_U32 WRITE_BUFFER_SIZE;

  /*Used to determine if and how much padding is necessary*/
	AEP_U32 padding;
	
	AEP_U32 hash_value;

	
	/*Loop forever*/
	while(1)
	{

		/*Read a message from the key manager*/
		rv = AEP_I_ReadFromSocket(keydb_fd, (AEP_U8*)&message_hdr, HDR_SIZE);
		if (rv != AEP_R_OK)
		{
			keydb_fd = -1;
			pthread_exit(NULL);
		}
		
	  /* Has this request just come through the key manager and is on it's way
	  to the accelerator.  The message could also originate in the key manager*/
		if (message_hdr.status == AEP_S_KEYMANAGER)
		{
			printf("Going down!!!\n");
		  /*Determine which request buffer this request is going to be added to and set
			  the variables up appropriately*/
			switch(message_hdr.flags & AEP_F_LUNMASK)
			{
			case AEP_F_ASIC :
				AEP_RequestBuffer		=	ASIC_CurRequestBuffer;
				AEP_RequestBuffer_mutex	=	&ASIC_RequestBuffer_mutex;
				send_now				=	&ASIC_send_now;
				send_complete			=	&ASIC_send_complete;
				BuffIndex				=	&ASICBuffIndex;
				SendBuffer				=	&SendASICBuffer;
				WRITE_BUFFER_SIZE		=   ASIC_BufferSize;
				break;
	
			case AEP_F_DES :
				AEP_RequestBuffer		=	DES_CurRequestBuffer;
				AEP_RequestBuffer_mutex	=	&DES_RequestBuffer_mutex;
				send_now				=	&DES_send_now;
				send_complete			=	&DES_send_complete;
				BuffIndex				=	&DESBuffIndex;
				SendBuffer				=	&SendDESBuffer;
				WRITE_BUFFER_SIZE		=   DES_BufferSize;
				break;

			case AEP_F_SA :
				AEP_RequestBuffer		=	SA_CurRequestBuffer;
				AEP_RequestBuffer_mutex	=	&SA_RequestBuffer_mutex;
				send_now				=	&SA_send_now;
				send_complete			=	&SA_send_complete;
				BuffIndex				=	&SABuffIndex;
				SendBuffer				=	&SendSABuffer;
				WRITE_BUFFER_SIZE		=   SA_BufferSize;
				break;
			}
 
		  /*Determine if any padding is required*/
			padding = message_hdr.datalen % sizeof(AEP_U64);

		  /*If padding is not 0 then we need to add some*/
			if (padding)
			{
			  /*The amount needed is 8 - the value of padding.  i.e. if padding was 1 then
				the datalen must be just 1 byte over a quad word boundry, so we must add
				7 bytes of padding (8 - 1).*/
				padding = sizeof(AEP_U64) - padding;
			}
	
		/*Lock the request buffer mutex.*/
			_LOCK(AEP_RequestBuffer_mutex);
			
 		 /*If adding this request to the request buffer will overflow the buffer, then
 		 send it for processing*/
			if (*BuffIndex + message_hdr.datalen + HDR_SIZE + padding >= WRITE_BUFFER_SIZE)
			{
			  /*The buffer should be sent.  This bool is checked in the request buffer handler thread*/
				*SendBuffer = TRUE;

			  /*Signal the send buffer condition variable*/
				pthread_cond_signal(send_now);

			  /*The request buffer hndlr thread resets the buffer index back to 0 when finished sending,
			  effectively emptying the buffer.  So this comparision will fail when the thread is finished*/
				while(*BuffIndex + message_hdr.datalen + HDR_SIZE + padding >= WRITE_BUFFER_SIZE)/*BUFFER_SIZE)*/
				  /*Wait until the request buffer handler thread signals us*/
					pthread_cond_wait(send_complete, AEP_RequestBuffer_mutex);
			}
	
		  /*Copy the data to the target.  Note : Memcpy would not give the correct
			endianess over 64 bits*/
			
			BEndToLEndCopy((AEP_U8*)AEP_RequestBuffer + *BuffIndex,
			(AEP_U8*)&message_hdr, HDR_SIZE / sizeof(AEP_U32));
		

		  /*Increment BuffIndex the appropriate amount*/
			(*BuffIndex) += HDR_SIZE;

		  /*If the message contained any data, read it from the socket directly into the request buffer*/
			if (message_hdr.datalen != 0)
			{
			/*Read directly from the socket into the buf*/
				   	rv = AEP_I_ReadFromSocket(keydb_fd, 
						(AEP_U8*)AEP_RequestBuffer + *BuffIndex,
					 	message_hdr.datalen);
			
	  			if(rv != AEP_R_OK)
				{
					AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ptrKeyManHndlr", rv, "Error reading Data from the Key Manager");
					(*BuffIndex) -= HDR_SIZE;
					_UNLOCK(AEP_RequestBuffer_mutex);
				}
				else
				{
		
				  /*Increment BuffIndex by the amount of data read.*/
					(*BuffIndex) += message_hdr.datalen;	

				  /*If padding is not 0 then we need to add some*/
					if (padding)
					{
					  /*Increment BuffIndex by the amount padding.*/
						(*BuffIndex) += padding;
					}
				}
			}
/*Unlock request buffer mutex*/
			_UNLOCK(AEP_RequestBuffer_mutex);
		
		}
		/*Request is to go back up to an application*/
		else
		{
	
		  /*Get the hash value for this transaction*/
			hash_value = AEP_I_HASH(message_hdr.trans_hndl);
										
/*Lock the response table*/
			_LOCK(&response_item_mutex_array[hash_value]);
			
		  /*Find the response item in the response hash table*/	
			rv = AEP_I_HashGetEntry(AEP_I_response_table.response_table, message_hdr.trans_hndl, 
				(AEP_VOID_PTR) &pResponse);
							
		  /*Check to see if the response item exists*/
			if (rv != AEP_R_OK) 
			{		
			 	 /*If not, discard the response*/
				AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ptrKeyManHndlr", rv, "No Corresponding Response Item for Message from Key Manager");

			}		  
			else
			{
			  /*Copy the complete response into the response table*/

			  /*Free the existing msg header*/
				AEP_free(pResponse->msg);

			  /*Allocate enough memory for the response*/	
				pResponse->msg = (AEP_U64_PTR) AEP_malloc(HDR_SIZE + message_hdr.datalen);
				if (pResponse->msg == NULL)
				{
					printf("returned null trying to alloc %d bytes", HDR_SIZE +
					message_hdr.datalen);
					AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ptrKeyManHndlr", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure - Exiting...");
					exit(1);
				}
			  
			  /*Copy the data from the target.  Note : Memcpy would not give the correct
				endianess over 64 bits*/
				memcpy(pResponse->msg + i, ((AEP_U8*)&message_hdr + i),
					HDR_SIZE);
			
			  /*If there's any data, read it back*/
				if (message_hdr.datalen != 0)
				{
					rv = AEP_I_ReadFromSocket(keydb_fd, (AEP_U8*)pResponse->msg + HDR_SIZE, 
					message_hdr.datalen);
					
					if (rv != AEP_R_OK)
						((AEP_MSG_HDR*)(pResponse->msg))->status = AEP_S_NOKEYMANAGER;
				
				}						
			  /*If there's a valid semaphore (ie someones waiting on this transaction)
				then signal them to let them know the answer has arrived*/
				if(pResponse->response_semaphore != NULL)
					sem_post(pResponse->response_semaphore);
			}
/*Unlock the response table mutex*/
			_UNLOCK(&response_item_mutex_array[hash_value]);
			
		}
	}
	
}
